import json
import logging
from threading import Lock
from typing import Any, List, Dict

from amplitude import Amplitude

from .config import LocalEvaluationConfig
from ..flagresult import FlagResult
from ..assignment import Assignment, AssignmentFilter, AssignmentService
from ..assignment.assignment_service import FLAG_TYPE_MUTUAL_EXCLUSION_GROUP, FLAG_TYPE_HOLDOUT_GROUP
from ..user import User
from ..connection_pool import HTTPConnectionPool
from .poller import Poller
from .evaluation.evaluation import evaluate
from ..variant import Variant
from ..version import __version__


class LocalEvaluationClient:
    """Experiment client for evaluating variants for a user locally."""

    def __init__(self, api_key: str, config: LocalEvaluationConfig = None):
        """
        Creates a new Experiment LocalEvaluationClient instance.
            Parameters:
                api_key (str): The environment API Key
                config (LocalEvaluationConfig): Config Object

            Returns:
                Experiment Client instance.
        """

        if not api_key:
            raise ValueError("Experiment API key is empty")
        self.api_key = api_key
        self.config = config or LocalEvaluationConfig()
        self.assignment_service = None
        if config and config.assignment_config:
            instance = Amplitude(config.assignment_config.api_key, config.assignment_config)
            self.assignment_service = AssignmentService(instance, AssignmentFilter(
                config.assignment_config.cache_capacity))
        self.logger = logging.getLogger("Amplitude")
        self.logger.addHandler(logging.StreamHandler())
        if self.config.debug:
            self.logger.setLevel(logging.DEBUG)
        self.__setup_connection_pool()
        self.flags = None
        self.poller = Poller(self.config.flag_config_polling_interval_millis / 1000, self.__do_flags)
        self.lock = Lock()

    def start(self):
        """
        Fetch initial flag configurations and start polling for updates. You must call this function to begin
        polling for flag config updates.
        """
        self.__do_flags()
        self.poller.start()

    def evaluate(self, user: User, flag_keys: List[str] = None) -> Dict[str, Variant]:
        """
         Locally evaluates flag variants for a user.
         Parameters:
                user (User): The user to evaluate
                flag_keys (List[str]): The flags to evaluate with the user. If empty, all flags from the flag cache are evaluated.

            Returns:
                The evaluated variants.
        """
        variants = {}
        if self.flags is None or len(self.flags) == 0:
            return variants
        user_json = str(user)
        self.logger.debug(f"[Experiment] Evaluate: User: {user_json} - Flags: {self.flags}")
        result_json = evaluate(self.flags, user_json)
        self.logger.debug(f"[Experiment] Evaluate Result: {result_json}")
        evaluation_result = json.loads(result_json)
        filter_result = flag_keys is not None
        assignment_result = {}
        for key, value in evaluation_result.items():
            included = not filter_result or key in flag_keys
            if not value.get('isDefaultVariant') and included:
                variants[key] = Variant(value['variant'].get('key'), value['variant'].get('payload'))
            if included or evaluation_result[key]['type'] == FLAG_TYPE_MUTUAL_EXCLUSION_GROUP or \
                    evaluation_result[key]['type'] == FLAG_TYPE_HOLDOUT_GROUP:
                assignment_result[key] = FlagResult(value)
        if self.assignment_service:
            self.assignment_service.track(Assignment(user, assignment_result))
        return variants

    def __do_flags(self):
        conn = self._connection_pool.acquire()
        headers = {
            'Authorization': f"Api-Key {self.api_key}",
            'Content-Type': 'application/json;charset=utf-8',
            'X-Amp-Exp-Library': f"experiment-python-server/{__version__}"
        }
        body = None
        self.logger.debug('[Experiment] Get flag configs')
        try:
            response = conn.request('GET', '/sdk/v1/flags', body, headers)
            response_body = response.read().decode("utf8")
            if response.status != 200:
                raise Exception(
                    f"[Experiment] Get flagConfigs - received error response: ${response.status}: ${response_body}")
            self.logger.debug(f"[Experiment] Got flag configs: {response_body}")
            self.lock.acquire()
            self.flags = response_body
            self.lock.release()
        finally:
            self._connection_pool.release(conn)

    def __setup_connection_pool(self):
        scheme, _, host = self.config.server_url.split('/', 3)
        timeout = self.config.flag_config_poller_request_timeout_millis / 1000
        self._connection_pool = HTTPConnectionPool(host, max_size=1, idle_timeout=30,
                                                   read_timeout=timeout, scheme=scheme)

    def stop(self) -> None:
        """
        Stop polling for flag configurations. Close resource like connection pool with client
        """
        self.poller.stop()
        self._connection_pool.close()

    def __enter__(self) -> 'LocalEvaluationClient':
        return self

    def __exit__(self, *exit_info: Any) -> None:
        self.stop()



class FlagResult:
    def __init__(self, result):
        self.variant = result.get('variant')
        self.description = result.get('description')
        self.is_default_variant = result.get('isDefaultVariant')
        self.exp_key = result.get('exp_key')
        self.deployed = result.get('deployed')
        self.type = result.get('type')



from amplitude import Amplitude, BaseEvent
from ..assignment.assignment import Assignment
from ..assignment.assignment import DAY_MILLIS
from ..assignment.assignment_filter import AssignmentFilter
from ..util import hash_code

FLAG_TYPE_MUTUAL_EXCLUSION_GROUP = "mutual-exclusion-group"
FLAG_TYPE_HOLDOUT_GROUP = "holdout-group"


def to_event(assignment: Assignment) -> BaseEvent:
    event = BaseEvent(event_type='[Experiment] Assignment', user_id=assignment.user.user_id,
                      device_id=assignment.user.device_id, event_properties={}, user_properties={})
    for key in sorted(assignment.results):
        event.event_properties[key + '.variant'] = assignment.results[key].variant['key']

    set_props = {}
    unset_props = {}

    for key in sorted(assignment.results):
        if assignment.results[key].type == FLAG_TYPE_MUTUAL_EXCLUSION_GROUP:
            continue
        elif assignment.results[key].is_default_variant:
            unset_props[f'[Experiment] {key}'] = '-'
        else:
            set_props[f'[Experiment] {key}'] = assignment.results[key].variant['key']

    event.user_properties['$set'] = set_props
    event.user_properties['$unset'] = unset_props

    event.insert_id = f'{event.user_id} {event.device_id} {hash_code(assignment.canonicalize())} {int(assignment.timestamp / DAY_MILLIS)}'

    return event


class AssignmentService:
    def __init__(self, amplitude: Amplitude, assignment_filter: AssignmentFilter):
        self.amplitude = amplitude
        self.assignmentFilter = assignment_filter

    def track(self, assignment: Assignment):
        if self.assignmentFilter.should_track(assignment):
            self.amplitude.track(to_event(assignment))



import time
from typing import Dict

from ..flagresult import FlagResult
from ..user import User

DAY_MILLIS = 24 * 60 * 60 * 1000


class Assignment:

    def __init__(self, user: User, results: Dict[str, FlagResult]):
        self.user = user
        self.results = results
        self.timestamp = time.time() * 1000

    def canonicalize(self) -> str:
        user = self.user.user_id.strip() if self.user.user_id else 'None'
        device = self.user.device_id.strip() if self.user.device_id else 'None'
        canonical = user + ' ' + device + ' '
        for key in sorted(self.results):
            value = self.results[key].variant['key'].strip() if self.results[key] and self.results[key].variant and \
                                                                self.results[key].variant.get('key') else 'None'
            canonical += key.strip() + ' ' + value + ' '
        return canonical
